#!/usr/bin/perl
#
# This example shows how to use the function test wrapper of AFS::Instant::Toolz::iaexec().
#
# -- Frank Burkhardt <burk@cbs.mpg.de>  Mon, 27 Feb 2012 12:52:24 +0100

use AFS::Instant::Toolz;
use IO::String;
use Data::Dumper;

=head FUNCTIONS

=over 4

=item $result=listcompare(\@a,\@b);

Compares two lists of strings. I<$result> == 1, if both lists are identical.

=cut

sub listcompare {
	my ($a,$b)=@_;
	if ( @{$a} != @{$b} ) {
		return 0;
	}
	for(my $n=0;$n<@{$a};$n++) {
		if ( $a->[$n] ne $b->[$n] ) {
			return 0;
		}
	}
	return 1;
}

=item $exitcode=iaexectesttest($prefix,\@command,$mode,$param)

This is an example iaexec()-replacement routine. It shows, how a replacement function has to work to be usable.
Possible values for I<$mode> are:

=over 4

=item B<readpipe>

I<$param> is an array ref into which the simulated output of the I<@command> is written. One array element
equals one line of output *without* newlines. Empty lines have to be empty strings.

=item B<writepipe>

I<$param> is a single string that has to be sent to the I<@command>.

=item B<readpipefh>

I<$param> is a scalar-reference to a perl-filehandle. A virtual filehandle has to be written into this scalar,
returning output of the fake I<@command> call.

=item B<none>

Output is written to stdout/stderr directly.

=back

=back

=cut

sub iaexectesttest {
	my ($prefix,$command,$mode,$param)=@_;
	if ( $prefix ne 'someprefix' ) {
		out('e',"iaexectesttest(): Something's wrong with my prefix");
		return 255;
	}
	
	# The test function only has to support expected commands...
	if ( listcompare($command,['echo','test']) ) {
		# ... and only the expected modes
		if ( $mode eq 'readpipe' ) {
			@{$param}=('test');
		} elsif ( $mode eq 'readpipefh' ) {
			my $output="test\n";
			my $io=IO::String->new($output);
			$$param=$io;
		} elsif ( $mode eq 'none' ) {
			print "test\n";
		} elsif ( $mode eq 'writepipe' ) {
			# echo doesn't support writing to it
			return 255;
		}
	} else {
		out('e',"Only the 'echo test' command is supported");
	}
	# echo test should always be successfull
	return 0;
}

sub function_to_be_tested {
	my @output;
	iaexec(
		["echo","test"],
		'readpipe',
		description => "Testing echo...",
		format => \@output,
	);
	if ( listcompare(\@output,['test']) ) {
		out('v',"Just what I expected");
	} else {
		out('e',"Something's wrong with the 'echo test' call");
		return 0;
	}
	if ( iaexec(
		["echo","test"],
		'test',
		description => "Testing echo...",
	) ) {
		out('v',"Nice, 'echo test' returned successfully");
	} else {
		out('e',"Something's wrong with the 'echo test' call");
		return 0;
	}
	if ( iaexec(
		["echo","test"],
		'exec',
		stdout => 'show',
		description => "Showing the string 'echo' by calling 'echo test'",
	) ) {
		out('v',"Nice, 'echo test' returned successfully once more.");
	} else {
		out('e',"Something's wrong with the 'echo test' call");
		return 0;
	}
}

# Prepareing test
iaexectest(\&iaexectesttest,['someprefix']);
$VERBOSE=2;
# Now testing the main function without any control over the iaexec()-call
exit ! function_to_be_tested();

