#!/usr/bin/perl -w
# Before `make install' is performed this script should be runnable with
# `make test'. After `make install' it should work as `perl AFS-Instant.t'


# change 'tests => 1' to 'tests => last_test_to_print';

#use Test::More tests => 14;
use Test::More;
use IO::String;
use Data::Dumper;

BEGIN { 
	use_ok('AFS::Instant::Toolz'); # tbt
	use_ok('AFS::Instant::CellDB'); # Tested
	use_ok('AFS::Instant::Update'); # Tested
	use_ok('AFS::Instant::Manage'); # partly tested
	use_ok('AFS::Instant::Output::Volumes'); # partly tested
	use_ok('AFS::Instant::Output::Partitions'); # partly tested
	use_ok('AFS::Instant::Distri'); # tbt
	use_ok('AFS::Instant::ServerUsers'); # tbt
	use_ok('AFS::Instant::aCall'); # tbt (partly)
	use_ok('AFS::Instant::Cred'); # tbt
};

use Fb::Data qw(setEquality human_readable setEquality listEquality align);
use Fb::IO qw(iaexec);
use Fb::Network;
use Fb::Out;
use AFS::Instant::Toolz;
use AFS::Instant::Manage;
use AFS::Instant::Update;
use AFS::Instant::CellDB;

use AFS::Instant::Output::Volumes;
use AFS::Instant::Output::Partitions;


# Insert your test code below, the Test::More module is use()ed here so read
# its man page ( perldoc Test::More ) for help writing this test script.



sub iaexec_fake_update {
	my ($command,$mode,$param)=@_;
	my $fh;
	if ( listEquality($command,['/usr/bin/vos','listvldb','-cell','example.com','-nosort','-noauth','-noresolve']) ) {
		open($fh,"< t/fake-output/vos_listvldb");
		$$param=$fh;
	} elsif ( listEquality($command,['/usr/bin/vos','listaddrs','-cell','example.com','-printuuid','-verbose','-noauth']) ) {
		open($fh,"< t/fake-output/vos_listaddr");
		$$param=$fh;
	} elsif ( listEquality($command,['/usr/bin/vos','listaddrs','-cell','example.com','-printuuid','-noresolve','-verbose','-noauth']) ) {
		open($fh,"< t/fake-output/vos_listaddr_noresolve");
		$$param=$fh;
	} elsif ( listEquality($command,['/usr/sbin/pt_util','-user']) ) {
		open($fh,"< t/fake-output/pt_util_user");
		$$param=$fh;
	} elsif ( listEquality($command,['/usr/sbin/pt_util','-members']) ) {
		open($fh,"< t/fake-output/pt_util_members");
		$$param=$fh;
	} elsif ( listEquality($command,['/usr/bin/timeout','-s','9','6s','/usr/bin/rxdebug','afsfs1.example.com','7000','-version']) ) {
		$$param=IO::String->new('');
	} elsif ( listEquality($command,['/usr/bin/timeout','-s','9','6s','/usr/bin/rxdebug','afsfs1-nic2.example.com','7000','-version']) ) {
		$$param=IO::String->new('');
		return 1 << 8;
	} elsif ( listEquality($command,['/usr/bin/timeout','-s','9','6s','/usr/bin/rxdebug','afsfs2.example.com','7000','-version']) ) {
		$$param=IO::String->new('');
	} elsif ( listEquality($command,['/usr/bin/timeout','-s','9','6s','/usr/bin/rxdebug','afsfs3.example.com','7000','-version']) ) {
		$$param=IO::String->new('');
	} elsif ( listEquality($command,['/usr/bin/timeout','-s','9','6s','/usr/bin/rxdebug','afsfs3-external.example.com','7000','-version']) ) {
		$$param=IO::String->new('');
		return 1 << 8;
	} elsif ( listEquality($command,['/usr/bin/timeout','-s','9','6s','/usr/bin/rxdebug','10.0.0.1','7000','-version']) ) {
		$$param=IO::String->new('');
	} elsif ( listEquality($command,['/usr/bin/timeout','-s','9','6s','/usr/bin/rxdebug','10.0.1.1','7000','-version']) ) {
		$$param=IO::String->new('');
		return 1 << 8;
	} elsif ( listEquality($command,['/usr/bin/timeout','-s','9','6s','/usr/bin/rxdebug','10.0.0.2','7000','-version']) ) {
		$$param=IO::String->new('');
	} elsif ( listEquality($command,['/usr/bin/timeout','-s','9','6s','/usr/bin/rxdebug','10.0.0.3','7000','-version']) ) {
		$$param=IO::String->new('');
	} elsif ( listEquality($command,['/usr/bin/timeout','-s','9','6s','/usr/bin/rxdebug','20.0.0.3','7000','-version']) ) {
		$$param=IO::String->new('');
		return 1 << 8;
	} elsif ( listEquality($command,['/usr/bin/vos','partinfo','10.0.0.1','-noauth']) ) {
		open($fh,"< t/fake-output/vos_partinfo_afsfs1");
		$$param=$fh;
	} elsif ( listEquality($command,['/usr/bin/vos','partinfo','10.0.0.2','-noauth']) ) {
		open($fh,"< t/fake-output/vos_partinfo_afsfs2");
		$$param=$fh;
	} elsif ( listEquality($command,['/usr/bin/vos','partinfo','10.0.0.3','-noauth']) ) {
		open($fh,"< t/fake-output/vos_partinfo_afsfs3");
		$$param=$fh;
	} elsif ( listEquality($command,['/usr/bin/vos','listvol','10.0.0.1','-format','-noresolve','-noauth']) ) {
		open($fh,"< t/fake-output/vos_listvol_afsfs1");
		$$param=$fh;
	} elsif ( listEquality($command,['/usr/bin/vos','listvol','10.0.0.2','-format','-noresolve','-noauth']) ) {
		open($fh,"< t/fake-output/vos_listvol_afsfs2");
		$$param=$fh;
	} elsif ( listEquality($command,['/usr/bin/vos','listvol','10.0.0.3','-format','-noresolve','-noauth']) ) {
		open($fh,"< t/fake-output/vos_listvol_afsfs3");
		$$param=$fh;
	} else {
		print STDERR Dumper(\@_);
		out('e',"Unknown iaexec() call");
		return undef;
	}
	$?=0;
	return 0;
}

my @last_commands=();


my $celldb;

sub serverInfo {
	my @serverlist=sort @{$celldb->serverInfo('uuidlist')};
	ok ( listEquality(\@serverlist,['0003d716-1501-1f3d-82-7a-0101007faa77','00085b10-4713-1d94-b1-83-4a40000aaa77','00835fe0-c594-1db6-92-0f-0b3a000aaa77']),"Server-UUID list");

	ok ( $celldb->serverInfo('uuid2ioaddress','0003d716-1501-1f3d-82-7a-0101007faa77') eq '10.0.0.1',"IO server address query");
	ok ( ! defined($celldb->serverInfo('uuid2ioaddress','wrong-uuid')),"negative IO server address query");

	ok ( $celldb->serverInfo('uuid2name','0003d716-1501-1f3d-82-7a-0101007faa77') eq 'afsfs1.example.com',"server UUID 2 name");
	ok ( listEquality($celldb->serverInfo('uuid2ioaddresses','0003d716-1501-1f3d-82-7a-0101007faa77'),['10.0.0.1']),"IO server address query");
	ok ( listEquality($celldb->serverInfo('uuid2addresses','0003d716-1501-1f3d-82-7a-0101007faa77'),['10.0.0.1','10.0.1.1']),"server addresses query");
	ok ( $celldb->serverInfo('addressonline','10.0.0.1'),"NIC online");
	ok ( ! $celldb->serverInfo('addressonline','10.0.1.1'),"negative NIC online test");
	ok ( ! defined($celldb->serverInfo('addressonline','10.0.2.1')),"non-existing NIC online test");
	ok ( $celldb->serverInfo('uuid2name','0003d716-1501-1f3d-82-7a-0101007faa77') eq 'afsfs1.example.com',"Server name query");
	ok ( $celldb->serverInfo('uuidonline','0003d716-1501-1f3d-82-7a-0101007faa77'),"Server online test");
	ok ( $celldb->serverInfo('address2uuid','10.0.0.1') eq '0003d716-1501-1f3d-82-7a-0101007faa77',"IP to UUID");
	ok ( $celldb->serverInfo('address2uuid','afsfs1') eq '0003d716-1501-1f3d-82-7a-0101007faa77',"IP to UUID with DNS lookup");
}

sub gettingVolumesAndInstances {
	# Requesting volumes and instances
	ok ( @{$celldb->getInstances('serveruuid' => '00085b10-4713-1d94-b1-83-4a40000aaa77',reference => 'servers')} == 3,"List of instances on afsfs2 via server");
	ok ( @{$celldb->getInstances('serveruuid' => '00085b10-4713-1d94-b1-83-4a40000aaa77',reference => 'all')} == 5,"List of instances on afsfs2 via server and VLDB");
	ok ( @{$celldb->getInstances('serveruuid' => '00085b10-4713-1d94-b1-83-4a40000aaa77',reference => 'vldb')} == 4,"List of instances on afsfs2 via VLDB");
	my $instance=$celldb->getInstance('www.readonly@afsfs3');
	ok ( ($instance->{'serveruuid'} eq '00835fe0-c594-1db6-92-0f-0b3a000aaa77') , "getInstance on a multi-RO volume");
	my $volref=$celldb->getVolume('root.cell');
	ok ( defined($volref),"Getting volume record record of 'root.cell'" );
}

sub celldbMisc {
	ok ( setEquality($celldb->listVolumes(),{'root.cell'=>1,'root.afs'=>1,'www'=>1,'www.temp'=>1,'broken'=>1,'norw'=>1}) , "Complete volume list");
	ok ( setEquality($celldb->listVolumes('prefix' => 'root.'),{'root.cell'=>1,'root.afs'=>1}) , "Prefix constrained volume list");
	ok ( listEquality($celldb->resolveVolumeId('102'),['root.afs',1]), "Resolving 102 to root.afs.readonly");
	ok ( ! defined($celldb->resolveVolumeId('1000')),"Resolving invalid volume instance ID failed");
	ok ( $celldb->isServerUuid('00085b10-4713-1d94-b1-83-4a40000aaa77'),"Valid serveruuid");
	ok ( ! $celldb->isServerUuid('invalid-uuid'),"Invalid serveruuid");
	ok ( $celldb->instanceTypeToId('readonly') == 1, "instanceTypeToId(readonly)");
	ok ( $celldb->instanceTypeToId(2) == 2,"instanceTypeToId(2)");
	ok ( ! defined($celldb->instanceTypeToId(3)),"instanceTypeToId(wrong_id)");
}

sub instanceInfo {
	my $volref=$celldb->getVolume('root.cell');
	# Testing volume reference
	ok ( $volref->{'rw'}->{'quota'} == 5120000,"Quota of 'root.cell' matching");
	ok ( @{$volref->{'ro'}} == 3,"Correct number of 'root.cell' RO-instances");
	ok ( $volref->{'rw'}->{'partsize'} == 2048000000000,"Partition size on 'root.cell' 's matching");
	ok ( $celldb->instanceInfo('storageleft',$volref->{'rw'}) == 4096000 , "Available storage for volume");
}

sub volumeInfo {
	my $volref=$celldb->getVolume('root.cell');
	ok ( $volref->{'reference'} eq 'all' ,"Correct default reference is in use");
	ok ( $volref->{'id'} eq '201',"ID of volume 'root.cell' is correct");
	ok ( $volref->{'roid'} eq '202',"RO-ID of volume 'root.cell'");
	ok ( $volref->{'backupid'} eq '203',"Backup-ID of volume 'root.cell'");
	ok ( $celldb->volumeInfo('exists','www'),"Volume 'www' exists" );
	ok ( $celldb->volumeInfo('rouptodate','root.afs') == 0,"root.afs is not uptodate");
	ok ( $celldb->volumeInfo('rouptodate',$volref) == 0,"root.cell has an unreleased RO and is therefor not uptodate");
	ok ( $celldb->volumeInfo('rouptodate','broken') ,"The RW might be offline but the ROs are uptodate");
	ok ( $celldb->volumeInfo('releasable','norw') == 0, "Volume without RW is not releasable");
	ok ( $celldb->volumeInfo('releasable',$volref) == 1 , "Perfectly releasable volume");

	ok ( ! $celldb->volumeInfo('backupuptodate','root.afs'), "Backup instance is not uptodate");
	ok ( ! $celldb->volumeInfo('backupuptodate','root.cell'), "No backup instance, not uptodate");
	ok ( $celldb->volumeInfo('backupuptodate','www.temp'), "Backup instance is uptodate");
	
}

sub ptdbInfo {
	ok ($celldb->ptdbInfo('id2name',1001) eq 'ernie',"PTDB object name resolving 1");
	ok ($celldb->ptdbInfo('id2name',-1000) eq 'users',"PTDB object name resolving 2");
	ok ( ! defined($celldb->ptdbInfo('id2name',1000)),"negative PTDB object name resolving");

	ok ($celldb->ptdbInfo('exists','ernie'),"PTDB object exists 1");
	ok ($celldb->ptdbInfo('exists','system:administrators'),"PTDB object exists 2");
	ok ( ! $celldb->ptdbInfo('exists','doesntexist'),"negative PTDB object existence test");

	ok ($celldb->ptdbInfo('name2id','anonymous') == 32766,"PTDB object ID resolving 1");
	ok ( $celldb->ptdbInfo('name2id','system:administrators') == -204,"PTDB object ID resolving 2");
	ok ( ! defined($celldb->ptdbInfo('name2id','doesntexist')),"negative PTDB object ID resolving");

	ok ( setEquality($celldb->ptdbInfo('hashid2name',{1001 => 2,1002 => 3,-1000 => 4}),{'ernie' => '2', 'bert' => '3','users' => 4}) ,"Bulk PTDB name resolving (implicit all)");
	ok ( setEquality($celldb->ptdbInfo('hashid2name',{1001 => 2,1002 => 3,-1000 => 4},include => 'all'),{'ernie' => '2', 'bert' => '3','users' => 4}) ,"Bulk PTDB name resolving (all)");
	ok ( setEquality($celldb->ptdbInfo('hashid2name',{1001 => 2,1002 => 3,-1000 => 4},include => 'groups'),{'users' => 4}) ,"Bulk PTDB name resolving (groups)");
	ok ( setEquality($celldb->ptdbInfo('hashid2name',{1001 => 2,1002 => 3,-1000 => 4},include => 'users'),{'ernie' => '2', 'bert' => '3'}) ,"Bulk PTDB name resolving (users)");

	ok ( $celldb->ptdbInfo('ismemberof','ernie',group => 'somegroup'),"Membership test");
	ok ( ! $celldb->ptdbInfo('ismemberof','bert',group => 'somegroup'),"negative membership test");
	ok ( $celldb->ptdbInfo('ismember+of','bert',group => 'somegroup'),"nested membership test");
	ok ( setEquality($celldb->ptdbInfo('list'),{'system:administrators' => 1,'system:backup' => 1,'system:ptsviewers' => 1,'system:authuser' => 1,'system:anyuser' => 1,'users' => 1,'somegroup' => 1,'somenestedgroup' => 1,'ernie' => 1,'bert' => 1,'anonymous' => 1,'admin' => 1}),"PTDB object list");
}

sub groupInfo {
	ok ( $celldb->groupInfo('exists','system:administrators'),"PTDB group exists");
	ok ( ! defined($celldb->groupInfo('exists','ernie')),"failed group existence test for a user");
	ok ( ! $celldb->groupInfo('exists','doesntexist'),"negative PTDB group existence test");

	ok ( $celldb->groupInfo('name2id','system:administrators') == -204,"PTDB group id resolving");
	ok ( ! defined($celldb->groupInfo('name2id','ernie')),"failed group ID resolving for a user");
	ok ( ! defined($celldb->groupInfo('name2id','doesntexist')),"negative PTDB group id resolving");

	ok ( $celldb->groupInfo('id2name',-1000) eq 'users',"PTDB group name resolving");
	ok ( ! defined($celldb->groupInfo('id2name',1001)),"failed group name resolving for a user");
	ok ( $celldb->groupInfo('id2name',-2000),"negative PTDB group id resolving");

	ok ( setEquality($celldb->groupInfo('members','somegroup'),{'ernie' => 1,'somenestedgroup' => 1}),"PTDB group member list (implicit all)");
	ok ( setEquality($celldb->groupInfo('members','somegroup',include => 'all'),{'ernie' => 1,'somenestedgroup' => 1}),"PTDB group member list (all)");
	ok ( setEquality($celldb->groupInfo('members','somegroup',include => 'groups'),{'somenestedgroup' => 1}),"PTDB group member list (groups only)");
	ok ( setEquality($celldb->groupInfo('members','somegroup',include => 'users'),{'ernie' => 1}),"PTDB group member list (users only)");
	ok ( ! defined($celldb->groupInfo('members','doesntexit')),"negative PTDB group member listing");
	ok ( ! defined($celldb->groupInfo('members','bert')),"failed group member listing for a user");

	ok ( scalar(keys %{$celldb->groupInfo('members+','somegroup')} == 3),"PTDB group member list (nested) 1");
	ok ( scalar(keys %{$celldb->groupInfo('members+','somenestedgroup')} == 1),"PTDB group member list (nested) 2");
	ok ( ! defined($celldb->groupInfo('members+','doesntexit')),"negative PTDB group member listing (nested)");
	ok ( ! defined($celldb->groupInfo('members+','bert')),"failed group membership listing (nested) for a user");

	ok ( setEquality($celldb->groupInfo('memberships','somenestedgroup'),{'somegroup' => []}),"PTDB group memberships 1");
	ok ( setEquality($celldb->groupInfo('memberships','somegroup'),{}),"PTDB group memberships 2");
	ok ( ! defined($celldb->groupInfo('memberships','doesntexist')),"negative PTDB group memberships listing");
	ok ( ! defined($celldb->groupInfo('memberships','ernie')),"failed group memberships listing for a user");

	ok ( setEquality($celldb->groupInfo('list'),{'system:administrators' => 1,'system:backup' => 1,'system:ptsviewers' => 1,'system:authuser' => 1,'system:anyuser' => 1,'users' => 1,'somegroup' => 1,'somenestedgroup' => 1}),"PTDB group list");
}

sub userInfo {
	ok ( $celldb->userInfo('exists','ernie'),"PTDB user exists");
	ok ( ! defined($celldb->userInfo('exists','users')),"failed user existence test for a group");
	ok ( ! $celldb->userInfo('exists','doesntexist'),"negative PTDB user existence test");

	ok ( ! $celldb->userInfo('ischangeable','admin'),"admin is a locked user");
	ok ( $celldb->userInfo('ischangeable','bert'),"bert is a regular user");
	ok ( $celldb->userInfo('ischangeable','doesntexit'),"Non-existent user can be changeable, too");
	
	ok ( $celldb->userInfo('name2id','bert') == 1002,"PTDB user name resolving");
	ok ( ! defined($celldb->userInfo('name2id','doesntexist')),"negative PTDB user name resolving");
	ok ( ! defined($celldb->userInfo('name2id','system:administrators')),"failed user name resolving for a group");
	
	ok ( $celldb->userInfo('id2name',1001) eq 'ernie',"PTDB user name resolving");
	ok ( ! defined($celldb->userInfo('id2name',-1000)),"failed user name2id for a user");
	ok ( ! defined($celldb->userInfo('id2name',2000)),"negative PTDB user ID resolving");
	
	ok ( setEquality($celldb->userInfo('memberships','ernie'),{'users' => 1,'somegroup' => 1}),"PTDB user memberships");
	ok ( ! defined($celldb->userInfo('memberships','somenestedgroup')),"failed user membership listing for a group");
	ok ( ! defined($celldb->userInfo('memberships','doesntexit')),"negative PTDB user membership listing");
	
	ok ( setEquality($celldb->userInfo('memberships+','bert'),{'users' => 1,'somegroup' => 1,'somenestedgroup' => 1}),"PTDB nested user memberships");
	ok ( ! defined($celldb->userInfo('memberships+','somenestedgroup')),"failed nested user membership listing for a group");
	ok ( ! defined($celldb->userInfo('memberships+','doesntexit')),"negative PTDB nested user membership listing");

	ok ( $celldb->userInfo('ismemberof','bert',group => 'users'),"PTDB user membership test");
	ok ( ! $celldb->userInfo('ismemberof','bert',group => 'somegroup'),"negative PTDB user membership test");
	ok ( ! defined($celldb->userInfo('ismemberof','somenestedgroup',group => 'somegroup')),"failed user membership test for a group");
	
	ok ( setEquality($celldb->userInfo('list'),{'ernie' => 1,'bert' => 1,'anonymous' => 1,'admin' => 1}),"PTDB user list");
}	

sub locations {
	ok ( setEquality($celldb->listVolumeGroups(),{'root' => 1,'all' => 1,'www' => 1}),"Volume group listing");
	my $lg=$celldb->getLocationGroup('internal-only',format => 'arrayref');
	ok ( @{$lg} == 3,"Three locations in location group 'internal-only'");
	my $cl=$celldb->expandLocationPattern('*/*',format => 'arrayref');
	ok ( @{$cl} == 4,"Four locations in the cell");

	my $partref=$celldb->getPartition('0003d716-1501-1f3d-82-7a-0101007faa77','a');
	ok ( $partref->{'partfree'} == 512000000,"getPartition got a partition");
	ok ( ! defined($celldb->getPartition('wrong-uid','a')),"negative getPartition");
	ok ( @{$celldb->getPartitions(format => 'arrayref')} == 4,"getPartitions got all partitions");
	
}

sub Volumematching {
	# Tests the checkVolumes() function which uses the volume matching infrastructure incl.
	# error labeling. One of the volumes ('norw') is locked an should fail the test.
	my %result=();
	my $testnode_=$celldb->{cellinfo}->{xc}->find("//ci:test[\@id='nolock']");
	my $result=$celldb->checkVolumes($testnode_->[0],{},result => \%result);
	ok ( $result == 4, "Volume test infrastructure (result,errorlevel)");
	if ( $result != -1 ) {
		ok ( setEquality($result{'comments'},{'norw' => 'Volume_Locked'}),"Volume test infrastructure (locked volume)");
		ok ( $result{'successcount'} == 5 ,"Volume test infrastructure (succeeding volumes)");
	}
	
	my @filters=(
		['root.*',['root.afs','root.cell'],"GLOB-Filter"],
		['globname=root.*',['root.afs','root.cell'],"GLOB-Filter (with keyword)"],
		['pcrename=(r..t\.afs|w{3})',['root.afs','www'],"PCRE-Filter"],
		
		['l=afsfs2.example.com/*:itype=all:vmode=servers',['root.afs','root.cell','www','broken','norw'],"Location (All types, no VLDB)"],
		['location=left-wing:vmode=vldb:itype=all',['root.afs','root.cell','www','broken'],"Location (All types, VLDB entries, VG)"],
		['location=afsfs1.example.com/a',['root.afs','root.cell','www','broken'],"Location (impl. RW, single part.)"],
		['location=external-accessible:match=any',["broken","norw","root.afs","root.cell","www","www.temp"],"Location (ANY-Match, VG)"],
		['location=afsfs1.example.com/a,external-accessible:itype=backup',['www.temp','root.afs'],"Location (Backup only, Multi-Expr.)"],

		['c=4',['root.cell','broken','www'],"ICount (default=all types)"],
		['count=4:itype=all',['root.cell','broken','www'],"ICount (all types explicitely)"],
		['count=3:itype=ro',['root.afs','root.cell','www','broken'],"Icount (RO)"],
		['count<2:itype=ro',['www.temp'],"Icount (LT-Operator, RO)"],
		['count<2:itype=ro:vmode=vldb',['www.temp','norw'],"Icount (LT-Operator, RO, VLDB entries)"],

		['f=locked',['norw'],"Flag: Locked volumes"],
		['flags=backupuptodate',['www.temp'],"Flag: backupuptodate"],
		['flags=releasable',['www','root.afs','root.cell'],"Flag: releasable"],
		['flags=rouptodate',['broken'],"Flag: RO up to date"],
		['flags=hasrwro',['root.afs','root.cell','www','broken'],"Flag: has RWRO"],
		['flags=hasoldro',['www'],"Flag: has old RO"],
		['flags=releasable,hasoldro',['root.afs','root.cell','www'],"Multiple flags with negation"],

		['e=needssalvage',['broken'],"Error: Salvage needed"],
		['errors=offline,partexist',['broken'],"Error: offline,partexit"],
		['errors=all',['broken','norw'],"Error: ALL"],
		
		['o=all',['www','root.cell','root.afs','norw','www.temp'],"IOnline: ALL"],
		['online!=all',['broken'],"IOnline: !ALL"],
		['online=2:itype=ro:vmode=vldb',['www','root.cell','broken'],"IOnline: RO"],		
		['online=2:itype=ro',['www','root.cell','broken'],"IOnline: RO"],

		['po=all',['www','root.cell','root.afs','norw','www.temp','broken'],"POnline: ALL"],
		['online!=all',['broken'],"POnline: !ALL"],
		
		['r>0',['www','root.cell','root.afs','norw','broken'],"SPOF"],

		# We'll use the file constrait to implicitely test, if the cond_check() function is working
		['files=200',['root.cell','root.afs','broken'],"Filecount (impl. RW)"],
		['root.cell files>=200:itype=rw',['root.cell'],"Filecount op test 1"],
		['root.cell !files=200:itype=rw',[],"Filecount op test 2"],
		['root.cell www files<=200:itype=rw',['root.cell'],"Filecount op test 3"],
		['root.cell !files<200:itype=rw',['root.cell'],"Filecount op test 4"],
		['root.cell !files>200:itype=rw',['root.cell'],"Filecount op test 5"],
		['files=250000:itype=ro',[],"Filecount (RO only, impl. all match)"],
		['files=250000:itype=ro:match=all',[],"Filecount (RO only, expl. all)"],
		['files=250000:itype=ro:match=any',['www'],"Filecount (RO only, expl. any)"],
		['quota>1200G',['www','www.temp'],"Quota (impl. RW)"],
		['size>900G',['www','www.temp'],"Volume size (impl. RW)"],
		['free>1G',[],"Free Space (Quota/Part)"],
		
		['s=all:onunrelased=0',['www.temp','broken'],"Sync (impl. RO,all)"],
		['sync=2:onunreleased=0',['root.cell'],"Sync (impl. RO)"],
		['sync=2:onunknown=0:onunreleased=0',['root.cell'],"Sync (impl. RO)"],
		['sync=2:onunknown=1:onunreleased=0',['www','norw'],"Sync (impl. RO)"],
		
		['lc=2',['norw'],"LCount (impl. RW/RO)"],
		['locationcount=2:vmode=vldb',[],"LCount (impl. RW/RO)"],
		['locationcount=1',['www.temp'],"LCount (impl. RW/RO)"],
	);
	for my $f (@filters) {
		my $filter=$celldb->parseVolumeFilter($f->[0]);
		ok ( defined($filter), "Parsing volume filter '$f->[0]' ($f->[2])" );
		if ( $filter ) {
			my $volumes=$celldb->filterVolumes($filter);
			my $compare=listEquality([sort keys %{$volumes}],[sort @{$f->[1]}]);
			ok ( $compare ,"Result of volume filter '$f->[0]' ($f->[2])");
			if ( ! $compare ) {
				print STDERR "Unexpected diff: test=[" . join(',',sort keys %{$volumes}) . "] expected=[" . join(',',sort @{$f->[1]}) . "]\n";
			}
		}
	}
}

sub Output {
	my $volref=$celldb->getVolume('root.cell');
	my $output;my $expected;my $timediff;my $voutput;my $poutput;
	
	
	# That one's tricky since the age of the volume changes...
	$voutput=AFS::Instant::Output::Volumes->new(celldb => $celldb,'hr' => 1,mode => 'bigtable',header => 1,nopref => 1);
	my $timediff_mtime=align(human_readable(time()-$volref->{'rw'}->{'mtime'},'timeinterval'),-4);

	$expected="Volumename               Size  Quota #I File  PFree Mtim RW      ROs                     Comment
root.cell              1000ki 5000ki  4  200  488Mi $timediff_mtime afs/a*  !afs/a+,afs/a*          -
";
	$output=$voutput->volumes({'root.cell' => $volref});
	ok ( $output eq $expected,"Volume output (HR)");
	if ( $output ne $expected ) {
		print STDERR Dumper($output,$expected);
	}

	# Output a modified bigtable style (via custom table)
	$voutput=AFS::Instant::Output::Volumes->new(celldb => $celldb,'hr' => 0,header => 1,
		'mode' => 'table:name|size|quota|instances|files|partfree|mtime|location|rolocations'
	);
	my $volumes=$celldb->listVolumes();
	for my $volume (keys %{$volumes}) {
		$volumes->{$volume}=$celldb->getVolume($volume);
	}
	$expected="Volumename Current_Size_RW Quota_RW Number_of_Instances Number_of_Inodes_RW Partfree_RW Mtime[RW] RW-Location Non-Cheap-RO-Locations
broken 1024000 5120000 4 200 512000000 1240000000 !afsfs1.example.com/a!* afsfs2.example.com/b*,afsfs3.example.com/a*
norw ? ? 3 ? ? ? - !afsfs2.example.com/b?,!afsfs3.example.com/a+
root.afs 1024000 5120000 5 200 512000000 1240000000 afsfs1.example.com/a+ afsfs2.example.com/a+,afsfs3.example.com/a+
root.cell 1024000 5120000 4 200 512000000 1240000000 afsfs1.example.com/a* !afsfs2.example.com/a+,afsfs3.example.com/a*
www 1024000000000 1536000000000 4 200000 512000000 1269000000 afsfs1.example.com/a* !afsfs2.example.com/a-,afsfs3.example.com/a-
www.temp 1024000000000 2048000000000 2 500000 512000000 1240000000 afsfs3.example.com/a 
";
	$output=$voutput->volumes($volumes);
	ok ( $output eq $expected,"Volume output (custom table)");

	$voutput=AFS::Instant::Output::Volumes->new(celldb => $celldb,header => 0,mode => 'name');
	$output=$voutput->volumes($volumes);
	$expected="broken
norw
root.afs
root.cell
www
www.temp
";
	ok( $output eq $expected,"Volume output (names only)");	
	$volref=$celldb->getVolume('www');
	my $partref=$celldb->getPartition('0003d716-1501-1f3d-82-7a-0101007faa77','a');

	$poutput=AFS::Instant::Output::Partitions->new(celldb => $celldb,'hr' => 0,'mode' => 'bigtable',nopref => 1);
	$timediff_mtime=time()-$volref->{'rw'}->{'mtime'};
	$timediff_infotime=time()-$volref->{'rw'}->{'infotime'};
	$expected="afsfs1.example.com/a 2048000000000 512000000 2047488000000 9 4 4 401600 200600 2048009216000 1024003072000 3072035840000 1536015360000 $timediff_mtime $timediff_infotime ?";
	ok ( $poutput->partition($partref) eq $expected,"Partition output");
	$poutput=AFS::Instant::Output::Partitions->new(celldb => $celldb,'hr' => 1,'mode' => 'bigtable',nopref => 1);
	$timediff_mtime=align(human_readable(time()-$volref->{'rw'}->{'mtime'},'timeinterval'),-4);
	$timediff_infotime=align(human_readable(time()-$volref->{'rw'}->{'infotime'},'timeinterval'),-4);
 	$expected="afs/a 1907Gi  488Mi 1906Gi     9     4     4  401k  200k 1907Gi  953Gi 2861Gi 1430Gi $timediff_mtime $timediff_infotime -";
	$output=$poutput->partition($partref);
	
	ok ( $output eq $expected,"Partition output (HR)");
	
	$poutput=AFS::Instant::Output::Partitions->new(celldb => $celldb,'hr' => 0,'mode' => 'name');
	$expected="afsfs1.example.com/a";
	$output=$poutput->partition($partref);
	ok ( $output eq $expected,"Partition output (names only)");
}

sub iaexecTests {
	my $iatt="t/iaexec-test";
	ok( iaexec([$iatt,"success"],'test'),"iaexec test success");
	ok( ! iaexec([$iatt,"error"],'test'),"iaexec test failure");
	
	ok( iaexec([$iatt,"success"],'exec'),"iaexec exec success");
	ok( ! iaexec([$iatt,"error"],'exec'),"iaexec exec failure");
	
	my @output;
	ok( iaexec([$iatt,"readpipe"],'readpipe',format => \@output),"iaexec readpipe success");
	ok( listEquality(\@output,['line1','line2']),"iaexec readpipe output");

	ok( ! iaexec([$iatt,"readpipe-error"],'readpipe',stdout => 'hide',format => \@output),"iaexec readpipe failure");
	ok( listEquality(\@output,['line1','line2']),"iaexec readpipe failure output");

	ok( ! iaexec([$iatt,"readpipe-error"],'readpipe',format => \@output,stdout => 'hide',stderr => 'include'),"iaexec readpipe failure with redir");
	ok( listEquality(\@output,['line1','fake_error_1','line2']),"iaexec readpipe failure with redir output");

	ok ( listEquality(iaexec([$iatt,"readpipe"],'readpipe',format => 'arrayref'),['line1','line2']),"iaexec readpipe arrayref ret");
	ok ( ! defined(iaexec([$iatt,"error"],'readpipe',format => 'arrayref')),"iaexec readpipe arrayref ret error");

	my %output;
	iaexec([$iatt,"readpipe"],'readpipe',format => \%output);
	ok( setEquality(\%output,{'line1' => 1,'line2' => 1}),"iaexec readpipe output as hash");
	ok ( setEquality(iaexec([$iatt,"readpipe"],'readpipe',format => 'hashref'),{'line1' => 1,'line2' => 1}),"iaexec readpipe hashref ret");
	my $output;
	iaexec([$iatt,"readpipe"],'readpipe',format => \$output);
	ok( $output eq "line1\nline2\n","iaexec readpipe output as string");
	ok ( iaexec([$iatt,"readpipe"],'readpipe',format => 'string') eq "line1\nline2\n","iaexec readpipe string ret");
	
	ok( iaexec([$iatt,"writepipe"],'writepipe',input => ["ok"]),"iaexec writepipe with correct input");
	ok( ! iaexec([$iatt,"writepipe"],'writepipe',input => ["notok"]),"iaexec writepipe with incorrect input");
}

sub iaexec_fake_manage {
	my ($command,$mode,$param)=@_;
	my $fh;
	if ( listEquality($command,['/usr/bin/vos','create','-server','10.0.0.1','-partition','a','-name','testvolume','-maxquota','1024','-cell','example.com']) ) {
		$$param=IO::String->new("Volume 701 created on partition /vicepa of 10.0.0.1\n");
	} elsif ( listEquality($command,['/usr/bin/vos','addsite','-server','10.0.0.1','-partition','a','-id','testvolume','-cell','example.com']) ) {
		$$param=IO::String->new("Added replication site 10.0.0.1 /vicepa for volume testvolume\n");
	} elsif ( listEquality($command,['/usr/bin/vos','remove','-server','10.0.0.1','-partition','a','-id','testvolume2','-cell','example.com']) ) {
		$$param=IO::String->new("Volume 701 on partition /vicepa server afsfs1.example.com deleted\n");
	} elsif ( listEquality($command,['/usr/bin/vos','remove','-server','10.0.0.1','-partition','a','-id','testvolume2.readonly','-cell','example.com']) ) {
		$$param=IO::String->new("Volume 702 on partition /vicepa server afsfs1.example.com deleted\n");
	} elsif ( listEquality($command,['/usr/bin/vos','rename','-oldname','testvolume','-newname','testvolume2','-cell','example.com']) ) {
		$$param=IO::String->new("Renamed volume testvolume to testvolume2\n");
	} elsif ( listEquality($command,['/usr/bin/vos','setfields','-id','testvolume2','-maxquota','2048','-cell','example.com']) ) {
		$$param=IO::String->new("");
	} elsif ( listEquality($command,['/usr/bin/vos','create','-server','10.0.0.1','-partition','a','-name','testvolume2','-maxquota','10240','-cell','example.com']) ) {
		$$param=IO::String->new("Volume 801 created on partition /vicepa of 10.0.0.1\n");
	} elsif ( listEquality($command,['/usr/bin/vos','backup','-id','testvolume2','-cell','example.com']) ) {
		$$param=IO::String->new("Created backup volume for testvolume2\n");
	} elsif ( listEquality($command,['/usr/bin/vos','addsite','-server','10.0.0.1','-partition','a','-id','testvolume2','-cell','example.com']) ) {
		$$param=IO::String->new("Added replication site 10.0.0.1 /vicepa for volume testvolume2\n");
	} elsif ( listEquality($command,['/usr/bin/vos','addsite','-server','10.0.0.2','-partition','b','-id','testvolume2','-cell','example.com']) ) {
		$$param=IO::String->new("Added replication site 10.0.0.1 /vicepa for volume testvolume2\n");
	} elsif ( listEquality($command,['/usr/bin/vos','release','-id','testvolume2','-cell','example.com']) ) {
		$$param=IO::String->new("Released volume testvolume2 successfully\n");
	} else {
		print STDERR Dumper(\@_);
		out('e',"Unknown iaexec() call");
		return undef;
	}
	push(@last_commands,$command);
	$?=0;
	return 0;
}


sub Manage {
	my $manage=AFS::Instant::Manage->new(celldb => $celldb);
	ok( $manage->createVolumeSimple('testvolume','0003d716-1501-1f3d-82-7a-0101007faa77','a',quota => 1024*1024),"Create volume + RW instance");
	ok( ( @last_commands == 1 ) && listEquality($last_commands[0],['/usr/bin/vos','create','-server','10.0.0.1','-partition','a','-name','testvolume','-maxquota','1024','-cell','example.com']),"Create volume + RW instance (commands)");
	@last_commands=();
	ok( $manage->createInstanceRo('testvolume',{'serveruuid' => '0003d716-1501-1f3d-82-7a-0101007faa77','partition' => 'a'}),"Create RO instance");
	ok( ( @last_commands == 1 ) && listEquality($last_commands[0],['/usr/bin/vos','addsite','-server','10.0.0.1','-partition','a','-id','testvolume','-cell','example.com']),"Create RO instance (commands)");
	@last_commands=();
	
	my $volref=$celldb->getVolume('testvolume');
	ok( $volref->{'rw'}->{'id'} == 701 ,"Create volume + RW instance (hinting)");
	ok( $volref->{'ro'}->[0]->{'rostate'} == 2,"Create RO instance (hinting)");

	ok( $manage->renameVolume('testvolume','testvolume2'),"Rename volume");
	ok( ( @last_commands == 1 ) && listEquality($last_commands[0],['/usr/bin/vos','rename','-oldname','testvolume','-newname','testvolume2','-cell','example.com']),"Rename volume (commands)");
	@last_commands=();

	ok( $manage->setVolumeQuota('testvolume2',2048*1024),"Set volume quota");
	ok( ( @last_commands == 1 ) && listEquality($last_commands[0],['/usr/bin/vos','setfields','-id','testvolume2','-maxquota','2048','-cell','example.com']),"Set volume quota (commands)");
	@last_commands=();

	$volref=$celldb->getVolume('testvolume2');
	ok( $volref->{'rw'}->{'quota'} == 2048*1024,"Set volume quota (hinting)");
	ok( $manage->removeInstance($volref->{'ro'}->[0]),"Remove RO instance");
	ok( ( @last_commands == 1 ) && listEquality($last_commands[0],['/usr/bin/vos','remove','-server','10.0.0.1','-partition','a','-id','testvolume2.readonly','-cell','example.com']),"Remove RO instance (commands)");
	@last_commands=();
	
	ok( $manage->removeInstance($volref->{'rw'}),"Remove RW instance + volume record");
	ok( ( @last_commands == 1 ) && listEquality($last_commands[0],['/usr/bin/vos','remove','-server','10.0.0.1','-partition','a','-id','testvolume2','-cell','example.com']),"Remove RW instance (commands)");
	@last_commands=();

	ok( $manage->createVolume('testvolume2',{'serveruuid' => '0003d716-1501-1f3d-82-7a-0101007faa77','partition' => 'a'},
		backup => 1,
		rolocations => {
			'0003d716-1501-1f3d-82-7a-0101007faa77' => { 'a' => 1 },
			'00085b10-4713-1d94-b1-83-4a40000aaa77' => { 'b' => 1 },
		},
		quota => 10*1024*1024,
		terrors => { 'volhash' => 1},
		volhash => 0,
	),"Creating volume + RO + Backup");
	ok( ( @last_commands == 4 ) &&
		listEquality($last_commands[0],['/usr/bin/vos','create','-server','10.0.0.1','-partition','a','-name','testvolume2','-maxquota','10240','-cell','example.com']) &&
		listEquality($last_commands[1],['/usr/bin/vos','backup','-id','testvolume2','-cell','example.com']) &&
		(
			(listEquality($last_commands[2],['/usr/bin/vos','addsite','-server','10.0.0.1','-partition','a','-id','testvolume2','-cell','example.com']) &&
			 listEquality($last_commands[3],['/usr/bin/vos','addsite','-server','10.0.0.2','-partition','b','-id','testvolume2','-cell','example.com']))
			or 
			(listEquality($last_commands[3],['/usr/bin/vos','addsite','-server','10.0.0.1','-partition','a','-id','testvolume2','-cell','example.com']) &&
			 listEquality($last_commands[2],['/usr/bin/vos','addsite','-server','10.0.0.2','-partition','b','-id','testvolume2','-cell','example.com']))
		)
	,'Creating volume + RO + Backup (commands)');
	@last_commands=();

	ok( $manage->syncVolumes(['testvolume2'],'readonly'),"Releasing volume");
	ok( ( @last_commands == 1 ) && listEquality($last_commands[0],['/usr/bin/vos','release','-id','testvolume2','-cell','example.com']),"Releasing volume (commands)");
	# Missing tests for group/user-management
}


sub Cred {
	my $cred=AFS::Instant::Cred->new(ftest => 1,ftest_basetime => 10,cellname => 'example.com',realmname => 'EXAMPLE.COM');
	# Before doing anything with the existing CC...
	ok( listEquality($cred->listTickets(),['afs/example.com@EXAMPLE.COM','host/somehost.somedomain@EXAMPLE.COM','krbtgt/EXAMPLE.COM@EXAMPLE.COM']),"Cred-Ticketlist untouched");
	ok( listEquality($cred->getOwner(),['exampleuser','EXAMPLE.COM']),"Cred-Owner untouched");
	ok( listEquality($cred->listTokens(),["example.com"]),"Cred-listTokens untouched");
	ok( $cred->getTgtTimeLeft() == 86340,"Cred-Tgttime untouched"); # 60 seconds into the 24h validity
	ok( $cred->getTgtTimeLeft(includerenew => 1) == 259140,"Cred-Tgttime incl. renew untouched");
	ok( $cred->getAfsTicketTimeLeft() == 86340,"Cred-afsticket untouched");
	ok( $cred->getTokenTimeLeft() == 86340,"Cred-tokentime untouched");
	$cred->renewTgt(); # Switching to 'renewed'
	ok( listEquality($cred->listTickets(),['afs/example.com@EXAMPLE.COM','krbtgt/EXAMPLE.COM@EXAMPLE.COM']),"Cred-Ticketlist renewed");
	ok( $cred->getTgtTimeLeft() == 86400,"Cred-Tgttime renewed"); # Directly at the point of renew
	ok( $cred->getTgtTimeLeft(includerenew => 1) == 259080,"Cred-Tgttime incl. renew renewed");
	ok( $cred->getAfsTicketTimeLeft() == 86400,"Cred-afsticket renewed");
	ok( $cred->getTokenTimeLeft() == 86400,"Cred-tokentime renewed");
	$cred->removeTickets(); # Switching to 'destroyed'
	ok( listEquality($cred->listTickets(),[]),"Cred-Ticketlist destroyed");
	ok( listEquality($cred->listTokens(),[]),"Cred-listTokens destroyed");
	ok( $cred->getTgtTimeLeft() == 0,"Cred-Tgttime destroyed");
	ok( $cred->getTgtTimeLeft(includerenew => 1) == 0,"Cred-Tgttime incl. renew destroyed");
	ok( $cred->getAfsTicketTimeLeft() == 0,"Cred-afsticket destroyed");
	ok( $cred->getTokenTimeLeft() == 0,"Cred-tokentime destroyed");
	$cred->fetchKerberosTgt('example2user',password => 'verysecret'); # Switching to 'new'
	ok( listEquality($cred->listTickets(),['afs/example.com@EXAMPLE.COM','krbtgt/EXAMPLE.COM@EXAMPLE.COM']),"Cred-Ticketlist new");
	ok( listEquality($cred->getOwner(),["example2user","EXAMPLE.COM"]),"Cred-Owner new");
	ok( $cred->getTgtTimeLeft() == 86400,"Cred-Tgttime new"); # Directly at the point of kinit
	ok( $cred->getTgtTimeLeft(includerenew => 1) == 259200,"Cred-Tgttime incl. renew new"); # Full renew-time for the first time in the test
	ok( $cred->getAfsTicketTimeLeft() == 86400,"Cred-afsticket new");
	ok( $cred->getTokenTimeLeft() == 86400,"Cred-tokentime new");
}

my %vdns=(
	'1' => ['afsfs1.example.com',0,2,0,'10.0.0.1'],
	'2' => ['afsfs2.example.com',0,2,0,'10.0.0.2'],
	'3' => ['afsfs3.example.com',0,2,0,'10.0.0.3'],
);
my %virtual_dns=(
	'afsfs1.example.com' => $vdns{1},'10.0.0.1' => $vdns{1}, 'afsfs1' => $vdns{1},
	'afsfs2.example.com' => $vdns{2},'10.0.0.2' => $vdns{2}, 'afsfs2' => $vdns{2},
	'afsfs3.example.com' => $vdns{3},'10.0.0.3' => $vdns{3}, 'afsfs3' => $vdns{3},
);

sub getHost_fake {
	my $name=shift;
	if ( defined($virtual_dns{$name}) ) {
		return @{$virtual_dns{$name}};
	} else {
		print STDERR "E: Invalid hostname: '$name'\n";
		return ();
	}
}

iaexecTests();



# Prepare virtual environment
Fb::IO::prepareFtests(iaexec => \&iaexec_fake_update);
AFS::Instant::Toolz::prepareFtests(cellname => 'example.com');
Fb::Network::prepareFtests(hostname => 'somehost.example.com',getHost => \&getHost_fake);

unlink("example.com.sqlite");
$celldb=AFS::Instant::CellDB->new(
	localdb => "./",
	localdbschema => "instantafs.sql",
	cellinfofile => "t/cell.xml",
	dnslookups => 0,
);
ok( defined($celldb), "CellDB-object connected to SQLite database");
my $update=AFS::Instant::Update->new(celldb => $celldb);
ok( defined($update), "Update-object ready");

# This is an integrated test of the Update-module since all of its components need to work
# to sync the virtual cell to the DB
my $updateresult=$update->run('*',ignoremissingptdb => 1,includeptdb => 1,parallel => 0);
ok( $updateresult,"Cell2DB sync test");
if ( ! $updateresult ) {
	exit 1;
}

serverInfo();
gettingVolumesAndInstances();
celldbMisc();
Cred();
instanceInfo();
volumeInfo();
ptdbInfo();
groupInfo();
userInfo();
locations();
Volumematching();
Output();

Fb::IO::prepareFtests(iaexec => \&iaexec_fake_manage);
Manage();

unlink("example.com.sqlite");
done_testing();
